export type ExpressMiddleware = import("express").RequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>> | import("express").ErrorRequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>>;
export type Logger = import("wootils/esm/node/logger").Logger;
export type ResponsesBuilder = import("../../services/http/responsesBuilder").ResponsesBuilder;
/**
 * <O>
 */
export type MiddlewareCreator<O> = (options: Partial<O>) => import("../../types").Middleware;
/**
 * Before reading the recevied information, these will be the settings for the response.
 */
export type ErrorHandlerDefaultOptions = {
    /**
     * The error message the response will show. Default `'Oops!
     * Something went wrong, please try again'`.
     */
    message: string;
    /**
     * The HTTP status code for the response. Default `500`.
     */
    status: number;
};
/**
 * The options for how to build the middleware responses.
 */
export type ErrorHandlerOptions = {
    /**
     * The options to build the default
     * response,
     * before the middleware analyzes the
     * recevied error.
     */
    default: ErrorHandlerDefaultOptions;
};
/**
 * @typedef {import('../../types').ExpressMiddleware} ExpressMiddleware
 * @typedef {import('../../types').Logger} Logger
 * @typedef {import('../../services/http/responsesBuilder').ResponsesBuilder} ResponsesBuilder
 * @prettierignore
 */
/**
 * @typedef {import('../../types').MiddlewareCreator<O>} MiddlewareCreator<O>
 * @template O
 */
/**
 * Before reading the recevied information, these will be the settings for the response.
 *
 * @typedef {Object} ErrorHandlerDefaultOptions
 * @property {string} message  The error message the response will show. Default `'Oops!
 *                             Something went wrong, please try again'`.
 * @property {number} status   The HTTP status code for the response. Default `500`.
 * @parent module:middlewares
 */
/**
 * The options for how to build the middleware responses.
 *
 * @typedef {Object} ErrorHandlerOptions
 * @property {ErrorHandlerDefaultOptions} default  The options to build the default
 *                                                 response,
 *                                                 before the middleware analyzes the
 *                                                 recevied error.
 * @parent module:middlewares
 */
/**
 * Provides the middleware to handle error responses for the app.
 *
 * @parent module:middlewares
 */
export class ErrorHandler {
    /**
     * @param {Logger}              appLogger         To log the received errors.
     * @param {ResponsesBuilder}    responsesBuilder  To generate the JSON response.
     * @param {boolean}             showErrors        If `false`, unknown errors will show a
     *                                                generic message instead of real
     *                                                message. And if `true`,
     *                                                it will not only show all kind of
     *                                                errors but it will also show the error
     *                                                stack.
     * @param {ClassAppError}       AppError          To validate if the received errors are
     *                                                known or not.
     * @param {ErrorHandlerOptions} [options={}]      Custom options to modify the
     *                                                middleware behavior.
     */
    constructor(appLogger: Logger, responsesBuilder: ResponsesBuilder, showErrors: boolean, AppError: any, options?: ErrorHandlerOptions);
    /**
     * A local reference for the `appLogger` service.
     *
     * @type {Logger}
     * @access protected
     * @ignore
     */
    _appLogger: Logger;
    /**
     * A local reference for the `responsesBuilder` service.
     *
     * @type {ResponsesBuilder}
     * @access protected
     * @ignore
     */
    _responsesBuilder: ResponsesBuilder;
    /**
     * Whether or not to show unknown errors real messages.
     *
     * @type {boolean}
     * @access protected
     * @ignore
     */
    _showErrors: boolean;
    /**
     * A local reference for the class the app uses to generate errors.
     *
     * @type {ClassAppError}
     * @access protected
     * @ignore
     */
    _AppError: any;
    /**
     * These are the "settings" the middleware will use in order to display the errors.
     *
     * @type {ErrorHandlerOptions}
     * @access protected
     * @ignore
     */
    _options: ErrorHandlerOptions;
    /**
     * Returns the Express middleware that shows the errors.
     *
     * @returns {ExpressMiddleware}
     */
    middleware(): ExpressMiddleware;
    /**
     * The options used to customize the middleware behavior.
     *
     * @returns {ErrorHandlerOptions}
     */
    get options(): ErrorHandlerOptions;
}
/**
 * Generates a middleware that show responses for unhandled errors thrown by the app.
 *
 * @type {MiddlewareCreator<ErrorHandlerOptions>}
 * @parent module:middlewares
 */
export const errorHandler: MiddlewareCreator<ErrorHandlerOptions>;
