export type ExpressMiddleware = import("express").RequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>> | import("express").ErrorRequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>>;
/**
 * <O>
 */
export type MiddlewareCreator<O> = (options: Partial<O>) => import("../../types").Middleware;
/**
 * The options to customize the HSTS header value.
 */
export type HSTSOptions = {
    /**
     * The time, in seconds, that the browser
     *          should remember that a site is only to be
     *          accessed using HTTPS.
     */
    maxAge?: number;
    /**
     * Whether or not the rule should apply to
     *   all sub domains.
     */
    includeSubDomains?: boolean;
    /**
     * Whether or not to include on the major
     *            browsers'
     *            preload list. This directive is not part
     *            of the specification, for more
     *            information about it, you should check
     *            the MDN documentation for the header.
     */
    preload?: boolean;
};
/**
 * @typedef {import('../../types').ExpressMiddleware} ExpressMiddleware
 */
/**
 * @typedef {import('../../types').MiddlewareCreator<O>} MiddlewareCreator<O>
 * @template O
 */
/**
 * The options to customize the HSTS header value.
 *
 * @typedef {Object} HSTSOptions
 * @property {number}  [maxAge=31536000]         The time, in seconds, that the browser
 *                                               should remember that a site is only to be
 *                                               accessed using HTTPS.
 * @property {boolean} [includeSubDomains=true]  Whether or not the rule should apply to
 *                                               all sub domains.
 * @property {boolean} [preload=false]           Whether or not to include on the major
 *                                               browsers'
 *                                               preload list. This directive is not part
 *                                               of the specification, for more
 *                                               information about it, you should check
 *                                               the MDN documentation for the header.
 * @see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Strict-Transport-Security
 * @parent module:middlewares
 */
/**
 * It configures a `Strict-Transport-Security` header and includes it on every response.
 *
 * @see https://tools.ietf.org/html/rfc6797
 * @parent module:middlewares
 */
export class HSTS {
    /**
     * @param {Partial<HSTSOptions>} [options={}]  The options to customize the header
     *                                             value.
     */
    constructor(options?: Partial<HSTSOptions>);
    /**
     * The value of the header that will be included on every response.
     *
     * @type {string}
     * @access protected
     * @ignore
     */
    _header: string;
    /**
     * Returns the Express middleware that includes the header on the response.
     *
     * @returns {ExpressMiddleware}
     */
    middleware(): ExpressMiddleware;
    /**
     * The value of the header that will be included on every response.
     *
     * @type {string}
     */
    get header(): string;
    /**
     * Creates the value for the header.
     *
     * @param {HSTSOptions} options  The options to customize the header value.
     * @returns {string}
     * @access protected
     * @ignore
     */
    _buildHeader(options: HSTSOptions): string;
}
/**
 * A middleware that creates and includes a `Strict-Transport-Security` header on every
 * response.
 * If the `options` parameter doesn't include the value for `maxAge`, the middleware will
 * ask the {@link AppConfiguration} service for the `hsts` settings as a fallback. You can
 * take advantage of that fallback to use the middleware and manage its settings from your
 * project configuration.
 * Both the `options` parameter and the `hsts` can include a `enabled` (boolean) flag to
 * either disable or enable the middleware.
 *
 * @type {MiddlewareCreator<HSTSOptions>}
 * @parent module:middlewares
 */
export const hsts: MiddlewareCreator<HSTSOptions>;
