export type ExpressMiddleware = import("express").RequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>> | import("express").ErrorRequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>>;
export type ExpressResponse = import("express").Response<any, Record<string, any>>;
export type ExpressNext = import("express").NextFunction;
export type EventsHub = import("wootils/esm/shared/eventsHub");
export type SendFile = (res: import("express").Response<any, Record<string, any>>, filepath: string, next: import("express").NextFunction, from?: string) => any;
export type HTMLGenerator = import("../../services/html/htmlGenerator").HTMLGenerator;
/**
 * <O>
 */
export type MiddlewareCreator<O> = (options: Partial<O>) => import("../../types").Middleware;
/**
 * The options to customize the behavior of the middleware.
 */
export type FastHTMLOptions = {
    /**
     * The name of the file the middleware will serve. It
     * can get overwritten if {@link FastHTML} receives an
     * {@link HTMLGenerator}, in that case, the file will
     * be obtained from that service. Default `index.html`.
     */
    file: string;
    /**
     * A list of regular expressions to match requests
     * paths that should be ignored.
     */
    ignore: RegExp[];
    /**
     * If `true`, {@link FastHTML} will get the list of all
     * routes controlled by {@link Jimpex} and will use
     * them to validate the incoming requests (in addition
     * to `ignore`): If a request URL doesn't match with
     * any of the controlled routes, it will show the HTML
     * file. Default `true`.
     */
    useAppRoutes: boolean;
};
export type FastHTMLMiddlewareOptions = FastHTMLOptions & FastHTMLMiddlewareOptionsProperties;
export type FastHTMLMiddlewareOptionsProperties = {
    /**
     * The name of a {@link HTMLGenerator} service for the middleware to use. Default
     * `htmlGenerator`.
     */
    htmlGenerator: string;
};
/**
 * @typedef {import('../../types').ExpressMiddleware} ExpressMiddleware
 * @typedef {import('../../types').ExpressResponse} ExpressResponse
 * @typedef {import('../../types').ExpressNext} ExpressNext
 * @typedef {import('../../types').EventsHub} EventsHub
 * @typedef {import('../../services/common/sendFile').SendFile} SendFile
 * @typedef {import('../../services/html/htmlGenerator').HTMLGenerator} HTMLGenerator
 */
/**
 * @typedef {import('../../types').MiddlewareCreator<O>} MiddlewareCreator<O>
 * @template O
 */
/**
 * The options to customize the behavior of the middleware.
 *
 * @typedef {Object} FastHTMLOptions
 * @property {string}   file          The name of the file the middleware will serve. It
 *                                    can get overwritten if {@link FastHTML} receives an
 *                                    {@link HTMLGenerator}, in that case, the file will
 *                                    be obtained from that service. Default `index.html`.
 * @property {RegExp[]} ignore        A list of regular expressions to match requests
 *                                    paths that should be ignored.
 * @property {boolean}  useAppRoutes  If `true`, {@link FastHTML} will get the list of all
 *                                    routes controlled by {@link Jimpex} and will use
 *                                    them to validate the incoming requests (in addition
 *                                    to `ignore`): If a request URL doesn't match with
 *                                    any of the controlled routes, it will show the HTML
 *                                    file. Default `true`.
 * @parent module:middlewares
 */
/**
 * @typedef {FastHTMLOptions & FastHTMLMiddlewareOptionsProperties} FastHTMLMiddlewareOptions
 * @prettierignore
 */
/**
 * @typedef {Object} FastHTMLMiddlewareOptionsProperties
 * @property {string} htmlGenerator
 * The name of a {@link HTMLGenerator} service for the middleware to use. Default
 * `htmlGenerator`.
 * @parent module:middlewares
 */
/**
 * It's common for an app to show an HTML view when no route was able to handle a request,
 * so the idea behind this middleware is to avoid going to every middleware and controller
 * and just specify that if the request is not for a route handled by a controller, just
 * serve the HTML and avoid processing unnecessary data.
 *
 * A simple example: The app has a route `/backend` that a frontend uses to get
 * information.
 * This middleware can be used to only allow the execution of middlewares and controllers
 * when the request route is for `/backend`.
 *
 * **Disclaimer**: Managing statics files with Express is not a best practice, but there
 * are scenarios where there is not other choice.
 *
 * @parent module:middlewares
 */
export class FastHTML {
    /**
     * @param {EventsHub}       events                To listen for the {@link Jimpex} event
     *                                                triggered after the app starts. The
     *                                                event is used to get all the
     *                                                controlled routes, in case the
     *                                                `useAppRoutes` option is set to true.
     * @param {SendFile}        sendFile              To send the HTML file response.
     * @param {FastHTMLOptions} [options={}]          To customize the middleware behavior.
     * @param {?HTMLGenerator}  [htmlGenerator=null]  If used, the file to serve will be the
     *                                                one generated by that service.
     */
    constructor(events: EventsHub, sendFile: SendFile, options?: FastHTMLOptions, htmlGenerator?: HTMLGenerator | null);
    /**
     * A local reference for the `events` service.
     *
     * @type {EventsHub}
     * @access protected
     * @ignore
     */
    _events: EventsHub;
    /**
     * A local reference for the `sendFile` service.
     *
     * @type {SendFile}
     * @access protected
     * @ignore
     */
    _sendFile: SendFile;
    /**
     * If specified, a reference for a service that generates HTML files.
     *
     * @type {?HTMLGenerator}
     * @access protected
     * @ignore
     */
    _htmlGenerator: HTMLGenerator | null;
    /**
     * The options that tell the middleware which routes should be ignored and which is
     * the file to serve.
     *
     * @type {FastHTMLOptions}
     */
    _options: FastHTMLOptions;
    /**
     * Whether or not the file is ready to be served, in case there's an
     * {@link HTMLGenerator}. If the service is used, the HTML is generated after the app
     * starts,
     * so the middleware will have to wait for it to be ready before being able to serve
     * it.
     *
     * @type {boolean}
     * @access protected
     * @ignore
     */
    _ready: boolean;
    /**
     * A list of regular expression that match the routes controlled by the app. This is
     * in case the `useAppRoutes` option is set to `true`; when the app gets started, an
     * event listener will obtain all the top controlled routes, create regular
     * expressions and save them on this property.
     *
     * @type {RegExp[]}
     * @access protected
     * @ignore
     */
    _routeExpressions: RegExp[];
    /**
     * Returns the Express middleware that validates the routes and serves the HTML file if
     * necessary.
     *
     * @returns {ExpressMiddleware}
     */
    middleware(): ExpressMiddleware;
    /**
     * The options that tell the middleware which routes should be ignored and which is the
     * file to serve.
     *
     * @type {FastHTMLOptions}
     * @todo Remove Object.freeze.
     */
    get options(): FastHTMLOptions;
    /**
     * Normalizes and validates the options recevied on the constructor.
     * If the class is using a {@link HTMLGenerator} service, the method will overwrite the
     * `file`
     * option with the result of the service's `getFile()` method.
     *
     * @param {FastHTMLOptions} options  The received options.
     * @returns {FastHTMLOptions}
     * @throws {Error} If no file and no {@link HTMLGenerator} service are specified.
     * @throws {Error} If no routes to ignore are specified and `useAppRoutes` is set to
     *                 `false`.
     * @access protected
     * @ignore
     */
    _normalizeOptions(options: FastHTMLOptions): FastHTMLOptions;
    /**
     * Serves the file on the response.
     *
     * @param {ExpressResponse} res   The server response.
     * @param {ExpressNext}     next  The function to call the next middleware.
     * @access protected
     * @ignore
     */
    _sendHTML(res: ExpressResponse, next: ExpressNext): void;
    /**
     * Adds the event listener that obtains the controlled routes when `useAppRoutes` is set
     * to `true`.
     *
     * @access protected
     * @ignore
     */
    _setupEvents(): void;
    /**
     * Checks whether a route should be ignored or not. The method checks first against the
     * `ignore`
     * option, and then against the controlled routes (if `useAppRoutes` is `false`, the
     * list will be empty).
     *
     * @param {string} route  The route to validate.
     * @returns {boolean}
     * @access protected
     * @ignore
     */
    _shouldIgnore(route: string): boolean;
}
/**
 * A middleware for filtering routes so you can serve an HTML before the app gets to
 * evaluate whether there's a controller for the requested route or not. For more
 * information about the reason of this middleware, please read the description of
 * {@link FastHTML}.
 *
 * @type {MiddlewareCreator<FastHTMLMiddlewareOptions>}
 * @parent module:middlewares
 */
export const fastHTML: MiddlewareCreator<FastHTMLMiddlewareOptions>;
