export type ExpressMiddleware = import("express").RequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>> | import("express").ErrorRequestHandler<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>>;
export type ExpressRequest = import("express").Request<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>>;
export type ResponsesBuilder = import("../../services/http/responsesBuilder").ResponsesBuilder;
/**
 * <O>
 */
export type MiddlewareCreator<O> = (options: Partial<O>) => import("../../types").Middleware;
/**
 * The options for how the middleware should behave if the requested version is `latest`.
 */
export type VersionValidatorLatestOptions = {
    /**
     * Whether or not the middleware should validate the _"latest
     * version"_.
     * Default `true`.
     */
    allow: boolean;
    /**
     * The name of the _"latest version"_. Basically,
     * `req.params.version`
     * must match with this property in order to be consider
     * "latest".
     * Default `'latest'`.
     */
    name: string;
};
/**
 * The options for how to detect if the request comes from a popup and how to compose the
 * post message the middleware will use to respond.
 */
export type VersionValidatorPopupOptions = {
    /**
     * The name of the query string variable the middleware will
     * check in order to indentify whether the request comes from
     * a popup or not.
     * The variable must have `'true'` as its value. Default
     * `'popup'`.
     */
    variable: string;
    /**
     * The title of the page that will be generated to respond in
     * case the versions don't match. Default `'Conflict'`.
     */
    title: string;
    /**
     * The contents of the post message the generated page will
     * send if the versions don't match. Default
     * `'vesion:conflict'`.
     */
    message: string;
};
/**
 * The options used to customize a {@link VersionValidator} instance.
 */
export type VersionValidatorOptions = {
    /**
     * The error message to show when the
     * version is invalid.
     */
    error: string;
    /**
     * The options for how the middleware
     * should behave if the requested
     * version is `latest`.
     */
    latest: VersionValidatorLatestOptions;
    /**
     * The options for how to detect if the
     * request comes from a popup and how
     * to compose the post message the
     * middleware will use to respond.
     */
    popup: VersionValidatorPopupOptions;
    /**
     * The version used to validate the
     * requests.
     * On the {@link VersionValidator}
     * constructor, if specified via
     * parameter,
     * the class will take care of
     * automatically add it to the options.
     */
    version: string | number;
};
/**
 * @typedef {import('../../types').ExpressMiddleware} ExpressMiddleware
 * @typedef {import('../../types').ExpressRequest} ExpressRequest
 * @typedef {import('../../services/http/responsesBuilder').ResponsesBuilder} ResponsesBuilder
 * @prettierignore
 */
/**
 * @typedef {import('../../types').MiddlewareCreator<O>} MiddlewareCreator<O>
 * @template O
 */
/**
 * The options for how the middleware should behave if the requested version is `latest`.
 *
 * @typedef {Object} VersionValidatorLatestOptions
 * @property {boolean} allow  Whether or not the middleware should validate the _"latest
 *                            version"_.
 *                            Default `true`.
 * @property {string}  name   The name of the _"latest version"_. Basically,
 *                            `req.params.version`
 *                            must match with this property in order to be consider
 *                            "latest".
 *                            Default `'latest'`.
 * @parent module:middlewares
 */
/**
 * The options for how to detect if the request comes from a popup and how to compose the
 * post message the middleware will use to respond.
 *
 * @typedef {Object} VersionValidatorPopupOptions
 * @property {string} variable  The name of the query string variable the middleware will
 *                              check in order to indentify whether the request comes from
 *                              a popup or not.
 *                              The variable must have `'true'` as its value. Default
 *                              `'popup'`.
 * @property {string} title     The title of the page that will be generated to respond in
 *                              case the versions don't match. Default `'Conflict'`.
 * @property {string} message   The contents of the post message the generated page will
 *                              send if the versions don't match. Default
 *                              `'vesion:conflict'`.
 * @parent module:middlewares
 */
/**
 * The options used to customize a {@link VersionValidator} instance.
 *
 * @typedef {Object} VersionValidatorOptions
 * @property {string}                        error    The error message to show when the
 *                                                    version is invalid.
 * @property {VersionValidatorLatestOptions} latest   The options for how the middleware
 *                                                    should behave if the requested
 *                                                    version is `latest`.
 * @property {VersionValidatorPopupOptions}  popup    The options for how to detect if the
 *                                                    request comes from a popup and how
 *                                                    to compose the post message the
 *                                                    middleware will use to respond.
 * @property {string | number}               version  The version used to validate the
 *                                                    requests.
 *                                                    On the {@link VersionValidator}
 *                                                    constructor, if specified via
 *                                                    parameter,
 *                                                    the class will take care of
 *                                                    automatically add it to the options.
 * @parent module:middlewares
 */
/**
 * This is the handler for the middleware/controller that validates the app version.
 * This is useful in cases where you want to restrict the access to the app to specific
 * versions,
 * for example: you have a frontend app which needs to be aligned with the "current"
 * version of the app, since the frontend won't realize a new version was released, the
 * validator can be used to let the frontend know.
 * Also, it can be configured to handle requests from popups, in which case, instead of
 * generating an error message, it will send a post message.
 *
 * @parent module:middlewares
 */
export class VersionValidator {
    /**
     * @param {?(string | ?number)} version
     * The current version of the app. The reason this is nullable is because this comes
     * directly from the app configuration, but you may want to re use this to validate
     * "another version", so you can use the custom shorthand and send the version using the
     * `options` parameter.
     * @param {ResponsesBuilder} responsesBuilder
     * To generate post message responses for popups.
     * @param {ClassAppError} AppError
     * To generate the error in case the version is invalid.
     * @param {Partial<VersionValidatorOptions>} [options={}]
     * Custom options to modify the middleware behavior.
     * @throws {Error}
     * If the version is `null` and the `options` don't include one either.
     */
    constructor(version: (string | (number | null)) | null, responsesBuilder: ResponsesBuilder, AppError: any, options?: Partial<VersionValidatorOptions>);
    /**
     * A local reference for the `responsesBuilder` service.
     *
     * @type {ResponsesBuilder}
     * @access protected
     * @ignore
     */
    _responsesBuilder: ResponsesBuilder;
    /**
     * A local reference for the class the app uses to generate errors.
     *
     * @type {ClassAppError}
     * @access protected
     * @ignore
     */
    _AppError: any;
    /**
     * These are the "settings" the middleware will use in order to validate the requests.
     *
     * @type {VersionValidatorOptions}
     * @access protected
     * @ignore
     */
    _options: VersionValidatorOptions;
    /**
     * Returns the Express middleware that will validate the `version` parameter.
     *
     * @returns {ExpressMiddleware}
     */
    middleware(): ExpressMiddleware;
    /**
     * The options used to customize the middleware behavior.
     *
     * @returns {VersionValidatorOptions}
     */
    get options(): VersionValidatorOptions;
    /**
     * Helper method that checks if the incoming request is from a popup.
     *
     * @param {ExpressRequest} req  The request information.
     * @returns {boolean}
     * @access protected
     * @ignore
     */
    _isPopup(req: ExpressRequest): boolean;
    /**
     * Helper method that checks if the "latest version" is enabled and if the given version
     * is "the latest" (comparing it with the option name).
     *
     * @param {string | number} version  The version to validate.
     * @returns {boolean}
     * @access protected
     * @ignore
     */
    _validateLatest(version: string | number): boolean;
}
/**
 * A middleware that will validate a `version` request parameter against the app version
 * and generate an error if they don't match.
 * This is a "middleware/controller" is because the wrappers for both are the same, the
 * difference is that, for controllers, Jimpex sends a second parameter with the route
 * where they are mounted.
 * By validating the route parameter, the function can know whether the implementation is
 * going to use the middleware by itself or as a route middleware.
 * If used as middleware, it will just return the result of
 * {@link VersionValidator#middleware};
 * but if used as controller, it will mount it on `[route]/:version/*`.
 *
 * @type {MiddlewareCreator<VersionValidatorOptions>}
 * @parent module:middlewares
 */
export const versionValidator: MiddlewareCreator<VersionValidatorOptions>;
