export type Provider = {
    register: import("wootils/esm/shared/jimpleFns").ProviderRegisterFn;
};
/**
 * @typedef {import('../../types').Provider} Provider
 */
/**
 * A simple subclass of `Error` but with support for context information.
 *
 * @augments Error
 * @parent module:services
 */
export class AppError extends Error {
    /**
     * @param {string} message       The error message.
     * @param {Object} [context={}]  Context information related to the error.
     */
    constructor(message: string, context?: any);
    /**
     * Context information related to the error.
     *
     * @type {Object}
     * @access protected
     * @todo Remove Object.freeze.
     */
    _context: any;
    /**
     * The date of when the error was generated.
     *
     * @type {Date}
     * @access protected
     */
    _date: Date;
    /**
     * Context information related to the error.
     *
     * @type {Object}
     */
    get context(): any;
    /**
     * The date of when the error was generated.
     *
     * @type {Date}
     */
    get date(): Date;
    /**
     * Information about the error that can be shown on an app response. This is set using
     * the `response` key on the `context`. The idea is that the error handler will read it
     * and use it on the response.
     *
     * @returns {Object}
     */
    get response(): any;
    /**
     * An HTTP status code related to the error. This is set using the `status` key on the
     * `context`. If the error handler finds it, it will use it as the response status,
     * and use it if necessary.
     *
     * @type {?number}
     */
    get status(): number;
    /**
     * Utility method that formats the context before saving it in the instance:
     * - If the context includes a `status` as a `string`, it will try to replace it with
     * its status code from the `statuses` package.
     *
     * @param {Object} original  The original context to format.
     * @returns {Object}
     * @access protected
     * @ignore
     */
    _parseContext(original: any): any;
}
/**
 * A service provider that will register both the {@link AppError} and a generator
 * function on the container. `AppError` will be the key for class, and `appError` will be
 * for the generator function.
 *
 * @type {Provider}
 * @example
 *
 *   // Register it on the container
 *   container.register(appError);
 *   // Getting access to the class.
 *   const AppError = container.get('AppError');
 *   // Getting access to the function.
 *   const appError = container.get('appError');
 *
 * @parent module:services
 */
export const appError: Provider;
