export type AppConfiguration = import("wootils/esm/node/appConfiguration").AppConfiguration;
export type Logger = import("wootils/esm/node/logger").Logger;
export type FrontendFs = import("../frontend/frontendFs").FrontendFs;
/**
 * <O>
 */
export type ProviderCreator<O> = (options?: Partial<O>) => import("wootils/esm/shared/jimpleFns").Provider;
/**
 * The options to customize the an `HTMLGenerator` service.
 */
export type HTMLGeneratorOptions = {
    /**
     * The name of the file it should use as
     * template.
     * Default `'index.tpl.html'`.
     */
    template: string;
    /**
     * The name of the generated file. Default
     * `'index.html'`.
     */
    file: string;
    /**
     * Whether or not to delete the tempalte after
     * generating the file. Default `true`.
     */
    deleteTemplateAfter: boolean;
    /**
     * The placeholder string where the information
     * will be written. Default
     * `'{{appConfiguration}}'`.
     */
    replacePlaceholder: string;
    /**
     * A regular expression for dynamic placeholders
     * that will be replaced by values when the file
     * is generated.
     * Default `'{{(.*?)}}'`.
     */
    valuesExpression: RegExp;
    /**
     * The name of the variable that will have the
     * information on the file. Default
     * `'appConfiguration'`.
     */
    variable: string;
    /**
     * A list of settings from the app configuration
     * that will be used as the information to
     * inject on the file.
     * Default `['features', 'version',
     * 'postMessagesPrefix']`.
     */
    configurationKeys: string[];
};
export type HTMLGeneratorProviderOptions = HTMLGeneratorOptions & HTMLGeneratorProviderOptionsProperties;
export type HTMLGeneratorProviderOptionsProperties = {
    /**
     * The name that will be used to register the service on the app container. This is to
     * allow multiple "instances" of the service to be created.
     */
    serviceName?: string;
    /**
     * The name of a service that the generator will use in order to read the values that will
     * be injected on the template. If the service is available, the values from
     * `configurationKeys`
     * will be ignored.
     */
    valuesService?: string | null;
};
export type HTMLGeneratorValuesServiceGetValues = () => Promise<any>;
/**
 * A service to provide the information value to an `HTMLGenerator` service to use on the
 * generated file.
 */
export type HTMLGeneratorValuesService = {
    /**
     * This is the method an `HTMLGenerator` service will call in order to retrieve the values
     * that should be injected on the generated file.
     */
    getValues: HTMLGeneratorValuesServiceGetValues;
};
/**
 * @typedef {import('../../types').AppConfiguration} AppConfiguration
 * @typedef {import('../../types').Logger} Logger
 * @typedef {import('../frontend/frontendFs').FrontendFs} FrontendFs
 */
/**
 * @typedef {import('../../types').ProviderCreator<O>} ProviderCreator<O>
 * @template O
 */
/**
 * The options to customize the an `HTMLGenerator` service.
 *
 * @typedef {Object} HTMLGeneratorOptions
 * @property {string}   template             The name of the file it should use as
 *                                           template.
 *                                           Default `'index.tpl.html'`.
 * @property {string}   file                 The name of the generated file. Default
 *                                           `'index.html'`.
 * @property {boolean}  deleteTemplateAfter  Whether or not to delete the tempalte after
 *                                           generating the file. Default `true`.
 * @property {string}   replacePlaceholder   The placeholder string where the information
 *                                           will be written. Default
 *                                           `'{{appConfiguration}}'`.
 * @property {RegExp}   valuesExpression     A regular expression for dynamic placeholders
 *                                           that will be replaced by values when the file
 *                                           is generated.
 *                                           Default `'{{(.*?)}}'`.
 * @property {string}   variable             The name of the variable that will have the
 *                                           information on the file. Default
 *                                           `'appConfiguration'`.
 * @property {string[]} configurationKeys    A list of settings from the app configuration
 *                                           that will be used as the information to
 *                                           inject on the file.
 *                                           Default `['features', 'version',
 *                                           'postMessagesPrefix']`.
 * @parent module:services
 */
/**
 * @typedef {HTMLGeneratorOptions & HTMLGeneratorProviderOptionsProperties} HTMLGeneratorProviderOptions
 * @parent module:services
 * @prettierignore
 */
/**
 * @typedef {Object} HTMLGeneratorProviderOptionsProperties
 * @property {string} [serviceName='htmlGenerator']
 * The name that will be used to register the service on the app container. This is to
 * allow multiple "instances" of the service to be created.
 * @property {?string} [valuesService='htmlGeneratorValues']
 * The name of a service that the generator will use in order to read the values that will
 * be injected on the template. If the service is available, the values from
 * `configurationKeys`
 * will be ignored.
 * @parent module:services
 */
/**
 * @callback HTMLGeneratorValuesServiceGetValues
 * @returns {Promise<Object>}
 * @parent module:services
 */
/**
 * A service to provide the information value to an `HTMLGenerator` service to use on the
 * generated file.
 *
 * @typedef {Object} HTMLGeneratorValuesService
 * @property {HTMLGeneratorValuesServiceGetValues} getValues
 * This is the method an `HTMLGenerator` service will call in order to retrieve the values
 * that should be injected on the generated file.
 * @parent module:services
 */
/**
 * This is a utility sever that generates an HTML file with custom information when the
 * app server is started.
 *
 * @parent module:services
 */
export class HTMLGenerator {
    /**
     * @param {AppConfiguration} appConfiguration
     * To read the values of the settings that are going to be send to the file.
     * @param {Logger} appLogger
     * To log messages when the file is generated, when the template is removed, and if it
     * happens,
     * when an error is thrown.
     * @param {FrontendFs} frontendFs
     * To read the contents of the template.
     * @param {Partial<HTMLGeneratorOptions>} [options={}]
     * To customize the service.
     * @param {?HTMLGeneratorValuesService} [valuesService=null]
     * If specified, instead of getting the values from the app configuration, they'll be
     * retrieved from this service `getValues` method.
     * @throws {Error}
     * If `valuesService` is specified but it doesn't have a `getValues` method.
     */
    constructor(appConfiguration: AppConfiguration, appLogger: Logger, frontendFs: FrontendFs, options?: Partial<HTMLGeneratorOptions>, valuesService?: HTMLGeneratorValuesService | null);
    /**
     * The service options.
     *
     * @type {HTMLGeneratorOptions}
     * @access protected
     * @ignore
     */
    _options: HTMLGeneratorOptions;
    /**
     * A local reference for the `appConfiguration` service.
     *
     * @type {AppConfiguration}
     * @access protected
     * @ignore
     */
    _appConfiguration: AppConfiguration;
    /**
     * A local reference for the `appLogger` service.
     *
     * @type {Logger}
     * @access protected
     * @ignore
     */
    _appLogger: Logger;
    /**
     * A local reference for the `frontendFs` service.
     *
     * @type {FrontendFs}
     * @access protected
     * @ignore
     */
    _frontendFs: FrontendFs;
    /**
     * A local reference for the recieved `valuesService` service.
     *
     * @type {?HTMLGeneratorValuesService}
     * @access protected
     * @ignore
     */
    _valuesService: HTMLGeneratorValuesService | null;
    /**
     * Whether or not the file has been generated.
     *
     * @type {boolean}
     * @access protected
     * @ignore
     */
    _fileReady: boolean;
    /**
     * A deferred promise to return when another service asks if the file has been
     * generated. Once this sevice finishes generating the file, the promise will be
     * resolved.
     *
     * @type {Object}
     * @access protected
     * @ignore
     */
    _fileDeferred: any;
    /**
     * Generate the HTML file.
     *
     * @returns {Promise}
     */
    generateHTML(): Promise<any>;
    /**
     * Get the name of the file the service generates.
     *
     * @returns {string}
     */
    getFile(): string;
    /**
     * Get the values that are going to be injected on the file.
     *
     * @returns {Promise<Object>}
     */
    getValues(): Promise<any>;
    /**
     * Returns a promise that will be resolved when the file has been generated.
     *
     * @returns {Promise}
     */
    whenReady(): Promise<any>;
    /**
     * The service options.
     *
     * @type {HTMLGeneratorOptions}
     * @todo Remove Object.freeze.
     */
    get options(): HTMLGeneratorOptions;
    /**
     * Get a value from an object dictionary using a string _"object path"_
     * (`prop.sub.otherProp`).
     * If the property doesn't exist or the path is invalid, it will return `null`.
     *
     * @param {Object} values     The dictionary from where the value will be read.
     * @param {string} valuePath  The path to the value.
     * @returns {*}
     * @access protected
     * @ignore
     * @todo Use ObjectUtils.get for this.
     */
    _getFromValues(values: any, valuePath: string): any;
    /**
     * Creates the code for the HTML file.
     *
     * @param {string} template  The template code where the values are going to be
     *                           injected.
     * @param {Object} values    The dictionary of values to inject.
     * @returns {string}
     * @access protected
     * @ignore
     */
    _processHTML(template: string, values: any): string;
}
/**
 * A service that hooks itself to the `after-start` event of the app server in order to
 * trigger the generation an the html file when the server starts.
 *
 * @type {ProviderCreator<HTMLGeneratorProviderOptions>}
 * @parent module:services
 */
export const htmlGenerator: ProviderCreator<HTMLGeneratorProviderOptions>;
