export type APIClientBase = import("wootils/esm/shared/apiClient");
export type APIClientEndpoints = {
    [x: string]: string | import("wootils/esm/shared/apiClient").APIClientEndpoint;
};
export type HTTP = import("./http").HTTP;
export type HTTPError = import("../common/httpError").HTTPError;
/**
 * <O>
 */
export type ProviderCreator<O> = (options?: Partial<O>) => import("wootils/esm/shared/jimpleFns").Provider;
/**
 * The options to customize how the service gets registered.
 */
export type APIClientProviderOptions = {
    /**
     * The name of the service that will be
     * registered into the app. Default
     * `'apiClient'`.
     */
    serviceName: string;
    /**
     * The name of the configuration
     * setting that has the API
     * information. Default `'api'`.
     */
    configurationSetting: string;
    /**
     * The class the service will
     * instantiate. It has to extend from
     * {@link APIClient}, which is the
     * default value.
     */
    clientClass: typeof APIClient;
};
/**
 * The configuration for the API the client will make requests to.
 */
export type APIClientConfiguration = {
    /**
     * The API entry point.
     */
    url: string;
    /**
     * A dictionary of named endpoints relative to
     * the API entry point.
     */
    endpoints: APIClientEndpoints;
};
/**
 * @typedef {import('../../types').APIClientBase} APIClientBase
 * @typedef {import('../../types').APIClientEndpoints} APIClientEndpoints
 * @typedef {import('./http').HTTP} HTTP
 * @typedef {import('../common/httpError').HTTPError} HTTPError
 */
/**
 * @typedef {import('../../types').ProviderCreator<O>} ProviderCreator<O>
 * @template O
 */
/**
 * The options to customize how the service gets registered.
 *
 * @typedef {Object} APIClientProviderOptions
 * @property {string}           serviceName           The name of the service that will be
 *                                                    registered into the app. Default
 *                                                    `'apiClient'`.
 * @property {string}           configurationSetting  The name of the configuration
 *                                                    setting that has the API
 *                                                    information. Default `'api'`.
 * @property {typeof APIClient} clientClass           The class the service will
 *                                                    instantiate. It has to extend from
 *                                                    {@link APIClient}, which is the
 *                                                    default value.
 * @parent module:services
 */
/**
 * The configuration for the API the client will make requests to.
 *
 * @typedef {Object} APIClientConfiguration
 * @property {string}             url        The API entry point.
 * @property {APIClientEndpoints} endpoints  A dictionary of named endpoints relative to
 *                                           the API entry point.
 * @parent module:services
 */
/**
 * An API client for the app to use. What makes this service special is that its that it
 * formats the received errors using the `AppError` service class and as fetch function it
 * uses the `http` service, allowing the app to to internally handle all the requests and
 * responses.
 *
 * @augments APIClientBase
 * @parent module:services
 */
export class APIClient extends APIClientBase {
    /**
     * @param {APIClientConfiguration} apiConfig  The configuration for the API the client
     *                                            will make requests to.
     * @param {HTTP}                   http       To get the `fetch` function for this
     *                                            service to use on all the requests.
     * @param {ClassHTTPError}         HTTPError  To format the received errors.
     */
    constructor(apiConfig: APIClientConfiguration, http: HTTP, HTTPError: any);
    /**
     * The configuration for the API the client will make requests to.
     *
     * @type {APIClientConfiguration}
     * @access protected
     * @ignore
     */
    _apiConfig: APIClientConfiguration;
    /**
     * A local reference for the class the app uses to generate HTTP errors.
     *
     * @type {ClassHTTPError}
     * @access protected
     * @ignore
     */
    _HTTPError: any;
    /**
     * Helper method that tries to get an error message from a given response.
     *
     * @param {Object} response                       A received response from a request.
     * @param {string} [fallback='Unexpected error']  A fallback message in case the method
     *                                                doesn't found one on the response.
     * @returns {string}
     */
    getErrorMessageFromResponse(response: any, fallback?: string): string;
    /**
     * The configuration for the API the client will make requests to.
     *
     * @type {APIClientConfiguration}
     * @todo Remove Object.freeze.
     */
    get apiConfig(): APIClientConfiguration;
}
/**
 * An API Client service to make requests to an API using endpoints defined on the app
 * configuration.
 *
 * @type {ProviderCreator<APIClientProviderOptions>}
 * @parent module:services
 */
export const apiClient: ProviderCreator<APIClientProviderOptions>;
import APIClientBase = require("wootils/shared/apiClient");
