export type Logger = import("wootils/esm/node/logger").Logger;
export type Provider = {
    register: import("wootils/esm/shared/jimpleFns").ProviderRegisterFn;
};
export type Response = any;
export type ExpressRequest = import("express").Request<import("express-serve-static-core").ParamsDictionary, any, any, qs.ParsedQs, Record<string, any>>;
export type HTTPFetchOptions = {
    /**
     * The request method.
     */
    method: string | null;
    /**
     * The request headers.
     */
    headers: {
        [x: string]: string;
    } | null;
    /**
     * The request body.
     */
    body: string | null;
    /**
     * The request query string parameters.
     */
    qs: {
        [x: string]: string;
    } | null;
    /**
     * An Express request object used to get
     * extra infromation (like headers and the
     * IP).
     */
    req: ExpressRequest | null;
};
/**
 * @typedef {import('../../types').Logger} Logger
 * @typedef {import('../../types').Provider} Provider
 * @typedef {import('node-fetch').Response} Response
 * @typedef {import('../../types').ExpressRequest} ExpressRequest
 */
/**
 * @typedef {Object} HTTPFetchOptions
 * @property {?string}                  method   The request method.
 * @property {?Object.<string, string>} headers  The request headers.
 * @property {?string}                  body     The request body.
 * @property {?Object.<string, string>} qs       The request query string parameters.
 * @property {?ExpressRequest}          req      An Express request object used to get
 *                                               extra infromation (like headers and the
 *                                               IP).
 * @parent module:services
 */
/**
 * A set of utilities to work with HTTP requests and responses.
 *
 * @parent module:services
 */
export class HTTP {
    /**
     * @param {boolean} logRequests  Whether or not to log the requests and their responses.
     * @param {Logger}  appLogger    If `logRequests` is `true`, this will be used to log
     *                               the requests and responses information.
     */
    constructor(logRequests: boolean, appLogger: Logger);
    /**
     * Whether or not to log the requests and their responses.
     *
     * @type {boolean}
     * @access protected
     * @ignore
     */
    _logRequests: boolean;
    /**
     * A local reference for the `appLogger` service.
     *
     * @type {Logger}
     * @access protected
     * @ignore
     */
    _appLogger: Logger;
    /**
     * Make a request.
     *
     * @param {string}                    url           The request URL.
     * @param {Partial<HTTPFetchOptions>} [options={}]  The request options.
     * @returns {Promise<Response>}
     */
    fetch(url: string, options?: Partial<HTTPFetchOptions>): Promise<Response>;
    /**
     * Creates a dictionary with all the custom headers a request has. By custom header it
     * means all the headers which name start with `x-`.
     * This method doesn't copy `x-forwarded-for` as the `fetch` method generates it by
     * calling `getIPFromRequest`.
     *
     * @param {ExpressRequest} req  The request from which it will try to get the headers.
     * @returns {Object.<string, string>}
     */
    getCustomHeadersFromRequest(req: ExpressRequest): {
        [x: string]: string;
    };
    /**
     * Try to get the IP from a given request.
     *
     * @param {ExpressRequest} req  The request from which it will try to obtain the IP
     *                              address.
     * @returns {?string}
     */
    getIPFromRequest(req: ExpressRequest): string | null;
    /**
     * It takes a dictionary of headers and normalize the names so each word will start with
     * an upper case character. This is helpful in case you added custom headers and didn't
     * care about the casing, or when copying headers from a server request, in which case
     * they are all tranformed to lower case.
     *
     * @param {Object.<string, string>} headers  The dictionary of headers to normalize.
     * @returns {Object.<string, string>}
     */
    normalizeHeaders(headers: {
        [x: string]: string;
    }): {
        [x: string]: string;
    };
    /**
     * Whether or not to log the requests and their responses.
     *
     * @type {boolean}
     */
    get logRequests(): boolean;
    /**
     * Log a a request information using the `appLogger` service.
     *
     * @param {string}           url      The request URL.
     * @param {HTTPFetchOptions} options  The options generated by the `fetch` method.
     * @access protected
     * @ignore
     */
    _logRequest(url: string, options: HTTPFetchOptions): void;
    /**
     * Log a a response information using the `appLogger` service.
     *
     * @param {Response} response  The response object returned by `node-fetch`.
     * @access protected
     * @ignore
     */
    _logResponse(response: any): void;
}
/**
 * The service provider that once registered on the app container will set an instance of
 * `HTTP` as the `http` service. The provider also checks the `debug.logRequests` setting
 * on the app configuration in order to enable or not the logging of requests.
 *
 * @type {Provider}
 * @example
 *
 *   // Register it on the container
 *   container.register(http);
 *   // Getting access to the service instance
 *   const http = container.get('http');
 *
 * @parent module:services
 */
export const http: Provider;
