export type Provider = {
    register: import("wootils/esm/shared/jimpleFns").ProviderRegisterFn;
};
export type AppConfiguration = import("wootils/esm/node/appConfiguration").AppConfiguration;
export type ExpressResponse = import("express").Response<any, Record<string, any>>;
/**
 * It allows customization of a post message HTML template.
 */
export type ResponsesBuilderPostMessageOptions = {
    /**
     * The target that will emit the `postMessage`.
     * Default `'window.opener'`.
     */
    target: string | null;
    /**
     * Whether or not to do a `window.close` after sending the message.
     * Default `true`.
     */
    close: boolean | null;
    /**
     * How many ms should it wait before closing the window, if `options.close` is `true`.
     * Default `700`.
     */
    closeDelay: number | null;
};
/**
 * @typedef {import('../../types').Provider} Provider
 * @typedef {import('../../types').AppConfiguration} AppConfiguration
 * @typedef {import('../../types').ExpressResponse} ExpressResponse
 */
/**
 * It allows customization of a post message HTML template.
 *
 * @typedef {Object} ResponsesBuilderPostMessageOptions
 * @property {?string} target
 * The target that will emit the `postMessage`.
 * Default `'window.opener'`.
 * @property {?boolean} close
 * Whether or not to do a `window.close` after sending the message.
 * Default `true`.
 * @property {?number} closeDelay
 * How many ms should it wait before closing the window, if `options.close` is `true`.
 * Default `700`.
 * @parent module:services
 */
/**
 * A utility service to build server responses.
 *
 * @parent module:services
 */
export class ResponsesBuilder {
    /**
     * @param {AppConfiguration} appConfiguration  To get the app version.
     */
    constructor(appConfiguration: AppConfiguration);
    /**
     * A local reference for the `appConfiguration` service.
     *
     * @type {AppConfiguration}
     * @access protected
     * @ignore
     */
    _appConfiguration: AppConfiguration;
    /**
     * Generates and send an HTML response that emits a post message.
     * The post message will be prefixed with the value of the configuration setting
     * `postMessagesPrefix`.
     *
     * @param {ExpressResponse} res
     * The Express response object necessary to write the HTML.
     * @param {string} title
     * The title for the HTML.
     * @param {string} message
     * The contents of the post message.
     * @param {number} [status=200]
     * The HTTP status.
     * @param {Partial<ResponsesBuilderPostMessageOptions>} [options={}]
     * Options to customize the HTML.
     */
    htmlPostMessage(res: ExpressResponse, title: string, message: string, status?: number, options?: Partial<ResponsesBuilderPostMessageOptions>): void;
    /**
     * Generates and sends a JSON response.
     *
     * @example
     *
     * <caption>The generated looks like this.</caption>
     *
     * {
     * metadata: {
     * version: '[app-version]',
     * status: [http-status],
     * },
     * data: [...],
     * }
     *
     * @param {ExpressResponse} res            The Express response object necessary to
     *                                         write the JSON.
     * @param {Object}          data           The information for the `data` key.
     * @param {number}          [status=200]   The HTTP status.
     * @param {Object}          [metadata={}]  Extra information to include on the
     *                                         `metadata` key.
     */
    json(res: ExpressResponse, data: any, status?: number, metadata?: any): void;
    /**
     * Generates a basic HTML template for other services to use.
     *
     * @param {string} title  The HTML `<title />` attribute.
     * @param {string} code   Javascript code to be wrapped on a `<script />` tag.
     * @returns {string}
     * @access protected
     * @ignore
     */
    _htmlTemplate(title: string, code: string): string;
    /**
     * Utility method used to make sure a recevied status is a valid status code. If the
     * status is a string, the method will try to find the code from the `statuses` package.
     *
     * @param {string | number} status  The status to normalize.
     * @returns {string | number} If `status` is a string, but there's no valid code, it
     *                            will return it as it was received.
     * @access protected
     * @ignore
     * @todo On the next breaking version, if there's no valid code, it will be
     *       transformed to `200`.
     */
    _normalizeStatus(status: string | number): string | number;
}
/**
 * The service provider that once registered on the app container will set an instance of
 * `ResponsesBuilder` as the `responsesBuilder` service.
 *
 * @type {Provider}
 * @example
 *
 *   // Register it on the container
 *   container.register(responsesBuilder);
 *   // Getting access to the service instance
 *   const responsesBuilder = container.get('responsesBuilder');
 *
 * @parent module:services
 */
export const responsesBuilder: Provider;
