/**
 * Given a server route definition, this function creates a regular expression to match
 * it: The expression replaces the routes parameters with placeholders so it can be
 * compared with real routes.
 *
 * @param {string}  route                  The route from which the expression will be
 *                                         created.
 * @param {boolean} [leadingSlash=true]    Whether or not the expression should match a
 *                                         leading slash.
 * @param {boolean} [trailingSlash=false]  Whether or not the expression should match a
 *                                         trailing slash. The reason this is `false` by
 *                                         default is because these expressions are often
 *                                         used to match against incoming requests, and
 *                                         they don't have a trailing slash.
 * @returns {RegExp}
 * @parent module:functions
 */
export function createRouteExpression(route: string, leadingSlash?: boolean, trailingSlash?: boolean): RegExp;
/**
 * Escapes a string to be used on `new RegExp(...)`.
 *
 * @param {string} text  The text to escape.
 * @returns {string}
 * @parent module:functions
 */
export function escapeForRegExp(text: string): string;
/**
 * @module functions
 */
/**
 * Removes any leading slash from a URL.
 *
 * @param {string} url  The URL to format.
 * @returns {string}
 * @parent module:functions
 */
export function removeLeadingSlash(url: string): string;
/**
 * Remove any leading and trailing slash from a URL.
 *
 * @param {string}  url              The URL to format.
 * @param {boolean} [leading=true]   Whether or not to remove any leading slash.
 * @param {boolean} [trailing=true]  Whether or not to remove the trailing slash.
 * @returns {string}
 * @parent module:functions
 */
export function removeSlashes(url: string, leading?: boolean, trailing?: boolean): string;
/**
 * Removes any trailing slash from a URL.
 *
 * @param {string} url  The URL to format.
 * @returns {string}
 * @parent module:functions
 */
export function removeTrailingSlash(url: string): string;
