export type Controller = {
    /**
     * The function Jimpex calls when mounting the
     * controller.
     */
    connect: import("../types").ControllerConnectFn;
    /**
     * A flag set to `true` to identify the
     * resource as a routes controller.
     */
    controller: boolean;
};
export type Middleware = {
    /**
     * The function Jimpex calls when mounting the
     * middleware.
     */
    connect: import("../types").MiddlewareConnectFn;
    /**
     * A flag set to `true` to identify the
     * resource as a middleware.
     */
    middleware: boolean;
};
export type ControllerConnectFn = (app: import("../app").Jimpex, route: string) => any;
export type ControllerCreatorFn = () => import("../types").ControllerConnectFn;
export type MiddlewareConnectFn = (app: import("../app").Jimpex) => import("../types").ExpressMiddleware;
export type MiddlewareCreatorFn = () => import("../types").MiddlewareConnectFn;
/**
 * <O>
 */
export type ControllerCreator<O> = (options: Partial<O>) => import("../types").Controller | import("../types").ControllerProvider;
/**
 * <O>
 */
export type MiddlewareCreator<O> = (options: Partial<O>) => import("../types").Middleware;
import { provider } from "wootils/shared/jimpleFns";
import { providerCreator } from "wootils/shared/jimpleFns";
import { providers } from "wootils/shared/jimpleFns";
/**
 * @typedef {import('../types').Controller} Controller
 * @typedef {import('../types').Middleware} Middleware
 * @typedef {import('../types').ControllerConnectFn} ControllerConnectFn
 * @typedef {import('../types').ControllerCreatorFn} ControllerCreatorFn
 * @typedef {import('../types').MiddlewareConnectFn} MiddlewareConnectFn
 * @typedef {import('../types').MiddlewareCreatorFn} MiddlewareCreatorFn
 */
/**
 * @typedef {import('../types').ControllerCreator<O>} ControllerCreator<O>
 * @template O
 */
/**
 * @typedef {import('../types').MiddlewareCreator<O>} MiddlewareCreator<O>
 * @template O
 */
/**
 * @module wrappers
 */
/**
 * Generates a routes controller for the application container to mount.
 *
 * @example
 *
 *   const myController = controller((app) => {
 *     const router = app.get('router');
 *     const ctrl = new MyController();
 *     return router.get('...', ctrl.doSomething()).post('...', ctrl.doSomethingElse());
 *   });
 *
 * @param {ControllerConnectFn} connectFn  A function that will be called the moment the
 *                                         app mounts the controller.
 * @returns {Controller}
 * @parent module:wrappers
 */
export function controller(connectFn: ControllerConnectFn): Controller;
/**
 * Generates a configurable routes controller for the application to mount. It's
 * configurable because the creator, instead of just being sent to the container to mount,
 * it can also be called as a function with custom parameters the controller can receive.
 *
 * @example
 *
 *   const myController = controllerCreator((options = {}) => (app) => {
 *     const router = app.get('router');
 *     const ctrl = new MyController(options);
 *     return router.get('...', ctrl.doSomething()).post('...', ctrl.doSomethingElse());
 *   });
 *
 * @param {ControllerCreatorFn} creatorFn  The function that generates the controller.
 * @returns {ControllerCreator<any>}
 * @parent module:wrappers
 */
export function controllerCreator(creatorFn: ControllerCreatorFn): ControllerCreator<any>;
/**
 * Generates a middleware for the app to use.
 *
 * @example
 *
 *   const myMiddleware = middleware((app) => (req, res, next) => {
 *     if (req.query.greet !== 'true') {
 *       return next();
 *     }
 *
 *     app.get('responsesBuilder').json(res, {
 *       hello: 'world',
 *     });
 *   });
 *
 * @example
 *
 * <caption>Disable the middleware base on a configuration option</caption>
 *
 *   const myMiddleware = middleware((app) => {
 *     if (app.get('appConfiguration').get('someOption') !== true) {
 *       return null;
 *     }
 *
 *     return (req, res, next) => {
 *       if (req.query.hello !== 'world') return next();
 *       app.get('responsesBuilder').json(res, {
 *         hello: 'world',
 *       });
 *     };
 *   });
 *
 * @param {MiddlewareConnectFn} connectFn  A function that will be called the moment the
 *                                         app mounts the middleware.
 * @returns {Middleware}
 * @parent module:wrappers
 */
export function middleware(connectFn: MiddlewareConnectFn): Middleware;
/**
 * Generates a configurable middleware for the app to use. It's configurable because the
 * creator,
 * instead of just being sent to the container to use, it can also be called as a function
 * with custom parameters the middleware can receive.
 *
 * @example
 *
 *   const myMiddleware = middlewareCreator((options = {}) => (app) =>
 *     options.enabled ? (req, res, next) => {} : null,
 *   );
 *
 * @param {MiddlewareCreatorFn} creatorFn  The function that generates the middleware.
 * @returns {MiddlewareCreator<any>}
 * @parent module:wrappers
 */
export function middlewareCreator(creatorFn: MiddlewareCreatorFn): MiddlewareCreator<any>;
export { provider, providerCreator, providers };
